// $Id: misc_support_addendum.cpp
// Copyright 2002  John George K.

// id3lib: a C++ library for creating and manipulating id3v1/v2 tags
// This library is free software; you can redistribute it and/or modify it
// under the terms of the GNU Library General Public License as published by
// the Free Software Foundation; either version 2 of the License, or (at your
// option) any later version.
//
// This library is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
// License for more details.
//
// You should have received a copy of the GNU Library General Public License
// along with this library; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

// The id3lib authors encourage improvements and optimisations to be sent to
// the id3lib coordinator.  Please see the README file for details on where to
// send such submissions.  See the AUTHORS file for a list of people who have
// contributed to id3lib.  See the ChangeLog file for a list of changes to
// id3lib.  

#if defined HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>

#include "misc_support_addendum.h"
#include "field.h"
#include "utils.h"

//using namespace dami;

ID3_Frame* ID3_AddGenre(ID3_Tag *tag, const char *text, bool replace)
{
  ID3_Frame* frame = NULL;
  if (NULL != tag && NULL != text && strlen(text) > 0)
  {
    if (replace)
    {
      ID3_RemoveGenres(tag);
    }
    if (replace || NULL == tag->Find(ID3FID_CONTENTTYPE))
    {
      frame = new ID3_Frame(ID3FID_CONTENTTYPE);
      if (NULL != frame)
      {
        frame->GetField(ID3FN_TEXT)->Set(text);
        tag->AttachFrame(frame);
      }
    }
  }
  
  return frame;
}

ID3_Frame* ID3_AddTrack(ID3_Tag *tag, const char *text, bool replace)
{
  ID3_Frame* frame = NULL;
  if (NULL != tag && NULL != text && strlen(text) > 0)
  {
    if (replace)
    {
      ID3_RemoveTracks(tag);
    }
    if (replace || NULL == tag->Find(ID3FID_TRACKNUM))
    {
      frame = new ID3_Frame(ID3FID_TRACKNUM);
      if (NULL != frame)
      {
        frame->GetField(ID3FN_TEXT)->Set(text);
        tag->AttachFrame(frame);
      }
    }
  }
  
  return frame;
}

int ID3_GetPictureData(const ID3_Tag *tag, const char *TempPicPath)
{
  if (NULL == tag) return 0;
  
  else
  {
    ID3_Frame *frame = NULL;
	frame = tag->Find(ID3FID_PICTURE);
	if (frame != NULL)
	{
		ID3_Field* myField = frame->GetField(ID3FN_DATA);
		if (myField != NULL)
		{
		  myField->ToFile(TempPicPath);
		  return (int)myField->Size();
		}
		else return 0;

	}
	else return 0;
  }
  
}

char *ID3_GetPictureMimeType(const ID3_Tag *tag)
{
  char *sPicMimetype = NULL;
  if (NULL == tag)
  {
    return sPicMimetype;
  }

  ID3_Frame *frame = NULL;
  frame = tag->Find(ID3FID_PICTURE);
  if (frame != NULL)
  {
    sPicMimetype = ID3_GetString(frame, ID3FN_MIMETYPE);
  }
  return sPicMimetype;
}

int ID3_HasPicture(const ID3_Tag *tag)
{
  if (NULL == tag) return 0;
  
  else
  {
    ID3_Frame *frame = tag->Find(ID3FID_PICTURE);
	if (frame != NULL)
	{
		ID3_Field* myField = frame->GetField(ID3FN_DATA);
		if (myField != NULL)
		{
		  return 1;
		}
		else return 0;

	}
	else return 0;
  }
  
}

ID3_Frame* ID3_AddPicture(ID3_Tag *tag, const char *TempPicPath, const char *MimeType, bool replace)
{
	ID3_Frame* frame = NULL;
  if (NULL != tag )
  {
    if (replace)
    {
      ID3_RemovePictures(tag);
    }
    if (replace || NULL == tag->Find(ID3FID_PICTURE))
    {
      frame = new ID3_Frame(ID3FID_PICTURE);
      if (NULL != frame)
      {
        frame->GetField(ID3FN_DATA)->FromFile(TempPicPath);
		frame->GetField(ID3FN_MIMETYPE)->Set(MimeType);
        tag->AttachFrame(frame);
      }
    }
  }
  return frame;
}

size_t ID3_RemovePictures(ID3_Tag *tag)
{
  size_t num_removed = 0;
  ID3_Frame *frame = NULL;

  if (NULL == tag)
  {
    return num_removed;
  }

  while ((frame = tag->Find(ID3FID_PICTURE)))
  {
    frame = tag->RemoveFrame(frame);
    delete frame;
    num_removed++;
  }

  return num_removed;
}

size_t ID3_RemovePictureType(ID3_Tag *tag, int pictype)
{
  size_t bremoved = 0;
  ID3_Frame *frame = NULL;

  if (NULL == tag)
  {
    return bremoved;
  }

  ID3_Tag::Iterator* iter = tag->CreateIterator();
   
   while (NULL != (frame = iter->GetNext()))
   {
    if(frame->GetID() == ID3FID_PICTURE)
	{
		if(frame->GetField(ID3FN_PICTURETYPE)->Get() == pictype) break;
	}
   }
   delete iter;
  
  if (NULL != frame)
  {
	frame = tag->RemoveFrame(frame);
   	delete frame;
   	bremoved = 1;
  }

  return bremoved;
}

ID3_Frame* ID3_AddPicture(ID3_Tag *tag, const char *TempPicPath, const char *MimeType, int pictype, const char* Description, bool replace)
{
	ID3_Frame* frame = NULL;
  if (NULL != tag )
  {
    if (replace)
    {
      ID3_RemovePictureType(tag, pictype);
    }
    if (replace || NULL == tag->Find(ID3FID_PICTURE))
    {
      frame = new ID3_Frame(ID3FID_PICTURE);
      if (NULL != frame)
      {
        frame->GetField(ID3FN_DATA)->FromFile(TempPicPath);
		frame->GetField(ID3FN_MIMETYPE)->Set(MimeType);
		frame->GetField(ID3FN_PICTURETYPE)->Set(pictype);
		frame->GetField(ID3FN_DESCRIPTION)->Set(Description);
        tag->AttachFrame(frame);
      }
    }
  }
  return frame;
}

int ID3_GetPictureDataOfPicType(ID3_Tag *tag, const char *TempPicPath, int pictype)
{
  if (NULL == tag) return 0;
  
  else
  {
    ID3_Frame *frame = NULL;
	ID3_Tag::Iterator* iter = tag->CreateIterator();
   
   while (NULL != (frame = iter->GetNext() ))
   {
    if(frame->GetID() == ID3FID_PICTURE)
	{
		if(frame->GetField(ID3FN_PICTURETYPE)->Get() == pictype) break;
	}
   }
   delete iter;
   
	if (frame != NULL)
	{
		ID3_Field* myField = frame->GetField(ID3FN_DATA);
		if (myField != NULL)
		{
		  myField->ToFile(TempPicPath);
		  return (int)myField->Size();
		}
		else return 0;

	}
	else return 0;
  }
  
}

char *ID3_GetMimeTypeOfPicType(ID3_Tag *tag, int pictype)
{
  char *sPicMimetype = NULL;
  if (NULL == tag)
  {
    return sPicMimetype;
  }

	ID3_Frame *frame = NULL;  
	ID3_Tag::Iterator* iter = tag->CreateIterator();
   
   while (NULL != (frame = iter->GetNext()))
   {
    if(frame->GetID() == ID3FID_PICTURE)
	{
		if(frame->GetField(ID3FN_PICTURETYPE)->Get() == pictype) break;
	}
   }
   delete iter;
	
  if (frame != NULL)
  {
    sPicMimetype = ID3_GetString(frame, ID3FN_MIMETYPE);
  }
  return sPicMimetype;
}

char *ID3_GetDescriptionOfPicType(ID3_Tag *tag, int pictype)
{
  char *sPicDescription = NULL;
  if (NULL == tag)
  {
    return sPicDescription;
  }

	ID3_Frame *frame = NULL;  
	ID3_Tag::Iterator* iter = tag->CreateIterator();
   
   while (NULL != (frame = iter->GetNext()))
   {
    if(frame->GetID() == ID3FID_PICTURE)
	{
		if(frame->GetField(ID3FN_PICTURETYPE)->Get() == pictype) break;
	}
   }
   delete iter;
	
  if (frame != NULL)
  {
    sPicDescription = ID3_GetString(frame, ID3FN_DESCRIPTION);
  }
  return sPicDescription;
}
